from pathlib import Path

import pytest
from fgpyo.sequence import reverse_complement
from pysam import FastxRecord

from pybwa import BwaIndex
from pybwa.libbwamem import BwaMem
from pybwa.libbwamem import BwaMemMode
from pybwa.libbwamem import BwaMemOptions


def test_bwamem_options_not_finalized() -> None:
    opt = BwaMemOptions().finalize()
    with pytest.raises(AttributeError):
        opt.min_seed_len = 2


def test_bwamem_options_set() -> None:
    opt = BwaMemOptions(
        min_seed_len=2,
        # mode: BwaMemMode | None = None,
        band_width=2,
        match_score=2,
        mismatch_penalty=2,
        minimum_score=2,
        unpaired_penalty=2,
        n_threads=2,
        skip_pairing=True,
        output_all_for_fragments=True,
        interleaved_paired_end=True,
        short_split_as_secondary=True,
        skip_mate_rescue=True,
        soft_clip_supplementary=True,
        with_xr_tag=True,
        query_coord_as_primary=True,
        keep_mapq_for_supplementary=True,
        with_xb_tag=True,
        max_occurrences=2,
        off_diagonal_x_dropoff=2,
        ignore_alternate_contigs=True,
        internal_seed_split_factor=2.1,
        drop_chain_fraction=2.1,
        max_mate_rescue_rounds=2,
        min_seeded_bases_in_chain=2,
        seed_occurrence_in_3rd_round=2,
        xa_max_hits=2,
        xa_drop_ratio=2.1,
        gap_open_penalty=2,
        gap_extension_penalty=2,
        clipping_penalty=2,
        threads=2,
        chunk_size=2,
    )

    # test the getters
    assert opt.min_seed_len == 2
    assert opt.band_width == 2
    assert opt.match_score == 2
    assert opt.mismatch_penalty == 2
    assert opt.minimum_score == 2
    assert opt.unpaired_penalty == 2
    assert opt.n_threads == 2
    assert opt.skip_pairing
    assert opt.output_all_for_fragments
    assert opt.interleaved_paired_end
    assert opt.short_split_as_secondary
    assert opt.skip_mate_rescue
    assert opt.soft_clip_supplementary
    assert opt.with_xr_tag
    assert opt.query_coord_as_primary
    assert opt.keep_mapq_for_supplementary
    assert opt.with_xb_tag
    assert opt.max_occurrences == 2
    assert opt.off_diagonal_x_dropoff == 2
    assert opt.ignore_alternate_contigs
    assert opt.internal_seed_split_factor == pytest.approx(2.1, 0.01)
    assert opt.drop_chain_fraction == pytest.approx(2.1, 0.01)
    assert opt.max_mate_rescue_rounds == 2
    assert opt.min_seeded_bases_in_chain == 2
    assert opt.seed_occurrence_in_3rd_round == 2
    assert opt.xa_max_hits == 2
    assert opt.xa_drop_ratio == pytest.approx(2.1, 0.01)
    assert opt.gap_open_penalty == 2
    assert opt.gap_extension_penalty == 2
    assert opt.clipping_penalty == 2
    assert opt.threads == 2
    assert opt.chunk_size == 2
    assert opt.mode is None

    # test that boolean flags work
    opt.skip_pairing = False
    assert not opt.skip_pairing
    opt.skip_pairing = True
    assert opt.skip_pairing

    # Test scaling

    # not scaled because it has been changed
    assert opt.gap_open_penalty == 2
    assert opt.gap_extension_penalty == 2
    assert not opt.finalized
    opt = opt.finalize()
    assert opt.finalized
    # even after finalize
    assert opt.gap_open_penalty == 2
    assert opt.gap_extension_penalty == 2


def test_bwa_mem_scaling() -> None:
    opt = BwaMemOptions(
        match_score=5,
    )
    # not scaled because it hasn't been finalized
    assert opt.gap_open_penalty == 6
    assert opt.gap_extension_penalty == 1
    opt = opt.finalize(copy=True)
    # scaled because it wasn't set
    assert opt.gap_open_penalty == 30
    assert opt.gap_extension_penalty == 5

    # test tuples
    opt = BwaMemOptions(clipping_penalty=(5, 3)).finalize()
    assert opt.clipping_penalty == (5, 3)


def test_bwamem_tuples_options() -> None:
    opt = BwaMemOptions(
        xa_max_hits=(2, 2),
        gap_open_penalty=(3, 3),
        gap_extension_penalty=(4, 4),
        clipping_penalty=(5, 5),
    )

    assert opt.xa_max_hits == 2
    assert opt.gap_open_penalty == 3
    assert opt.gap_extension_penalty == 4
    assert opt.clipping_penalty == 5

    opt = BwaMemOptions(
        xa_max_hits=(2, 3),
        gap_open_penalty=(4, 5),
        gap_extension_penalty=(6, 7),
        clipping_penalty=(8, 9),
    )
    assert opt.xa_max_hits == (2, 3)
    assert opt.gap_open_penalty == (4, 5)
    assert opt.gap_extension_penalty == (6, 7)
    assert opt.clipping_penalty == (8, 9)


def test_bwamem_options_mode() -> None:
    opt = BwaMemOptions(mode=BwaMemMode.INTRACTG).finalize()
    assert opt.clipping_penalty == 5

    opt = BwaMemOptions(mode=BwaMemMode.ONT2D).finalize()
    assert opt.clipping_penalty == 0
    assert opt.min_seeded_bases_in_chain == 20

    opt = BwaMemOptions(mode=BwaMemMode.PACBIO).finalize()
    assert opt.clipping_penalty == 0
    assert opt.min_seeded_bases_in_chain == 40


def test_bwamem(e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord) -> None:
    opt = BwaMemOptions(with_xr_tag=True)
    bwa = BwaMem(prefix=e_coli_k12_fasta)

    revcomp_seq = (
        None
        if not e_coli_k12_fastx_record.sequence
        else reverse_complement(e_coli_k12_fastx_record.sequence)
    )
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    recs_of_recs = bwa.align(opt=opt, queries=[e_coli_k12_fastx_record, revcomp_record])
    assert len(recs_of_recs) == 2

    assert len(recs_of_recs[0]) == 1
    rec = recs_of_recs[0][0]
    assert rec.query_name == "test"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_forward
    assert rec.cigarstring == "80M"

    assert len(recs_of_recs[1]) == 1
    rec = recs_of_recs[1][0]
    assert rec.query_name == "revcomp"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_reverse
    assert rec.cigarstring == "80M"
    # TODO: test multi-mapping, reverse strand, etc

    # NB: XA amd XB not generated for these records
    expected_tags = ["NM", "MD", "AS", "XS", "XR"]
    for recs in recs_of_recs:
        for rec in recs:
            for tag in expected_tags:
                assert rec.has_tag(tag), f"Missing tag {tag} in: {rec}"


def test_bwamem_from_index(e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord) -> None:
    index = BwaIndex(prefix=e_coli_k12_fasta)
    BwaMem(index=index)


def test_bwamem_no_index() -> None:
    with pytest.raises(ValueError):
        BwaMem()


def test_bwamem_align_no_options(
    e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord
) -> None:
    bwa = BwaMem(prefix=e_coli_k12_fasta)
    queries = [e_coli_k12_fastx_record]
    list_of_recs = bwa.align(queries=queries)
    assert len(list_of_recs) == 1
    assert len(list_of_recs[0]) == 1


def test_bwamem_align_no_queries(e_coli_k12_fasta: Path) -> None:
    bwa = BwaMem(prefix=e_coli_k12_fasta)
    list_of_recs = bwa.align(queries=[])
    assert len(list_of_recs) == 0


def test_bwamem_align_from_string(
    e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord
) -> None:
    opt = BwaMemOptions(threads=2)
    bwa = BwaMem(prefix=e_coli_k12_fasta)

    revcomp_seq = (
        None
        if not e_coli_k12_fastx_record.sequence
        else reverse_complement(e_coli_k12_fastx_record.sequence)
    )
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)
    assert e_coli_k12_fastx_record.sequence is not None
    assert revcomp_record.sequence is not None

    queries: list[str] = [
        e_coli_k12_fastx_record.sequence if i % 2 == 0 else revcomp_record.sequence
        for i in range(100)
    ]
    list_of_recs = bwa.align(opt=opt, queries=queries)
    assert len(list_of_recs) == len(queries)
    for i, recs in enumerate(list_of_recs):
        assert len(recs) == 1
        rec = recs[0]
        assert rec.query_name == f"read.{i}"
        if i % 2 == 0:
            assert rec.is_forward
        else:
            assert rec.is_reverse
        assert not rec.is_paired
        assert not rec.is_read1
        assert not rec.is_read2
        assert rec.reference_start == 80
        assert rec.cigarstring == "80M"


def test_bwamem_threading(e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord) -> None:
    opt = BwaMemOptions(threads=2)
    bwa = BwaMem(prefix=e_coli_k12_fasta)

    revcomp_seq = (
        None
        if not e_coli_k12_fastx_record.sequence
        else reverse_complement(e_coli_k12_fastx_record.sequence)
    )
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    queries = [e_coli_k12_fastx_record if i % 2 == 0 else revcomp_record for i in range(100)]
    list_of_recs = bwa.align(opt=opt, queries=queries)
    assert len(list_of_recs) == len(queries)
    for i, recs in enumerate(list_of_recs):
        assert len(recs) == 1
        rec = recs[0]
        if i % 2 == 0:
            assert rec.query_name == "test"
            assert rec.is_forward
        else:
            assert rec.query_name == "revcomp"
            assert rec.is_reverse
        assert not rec.is_paired
        assert not rec.is_read1
        assert not rec.is_read2
        assert rec.reference_start == 80
        assert rec.cigarstring == "80M"


def test_bwamem_align_multi_map(
    e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord
) -> None:
    opt = BwaMemOptions(
        min_seed_len=10,
        minimum_score=10,
        output_all_for_fragments=True,
        short_split_as_secondary=True,
    )
    bwa = BwaMem(prefix=e_coli_k12_fasta)

    assert e_coli_k12_fastx_record.sequence is not None
    query = FastxRecord(
        name=e_coli_k12_fastx_record.name,
        sequence=e_coli_k12_fastx_record.sequence[:10] + e_coli_k12_fastx_record.sequence[-10:],
    )
    list_of_recs = bwa.align(opt=opt, queries=[query])
    assert len(list_of_recs) == 1
    assert len(list_of_recs[0]) == 27
    rec = list_of_recs[0][0]
    assert rec.reference_start == 3618128
    assert rec.cigarstring == "14M6S"
    score = int(rec.get_tag("AS"))
    for rec in list_of_recs[0]:
        assert score >= int(rec.get_tag("AS"))


@pytest.mark.parametrize("query_coord_as_primary", [True])
def test_bwamem_align_sa_tag(e_coli_k12_fasta: Path, query_coord_as_primary: bool) -> None:
    # alignments below are defined based on (1) alignment score, then (2) genomic order
    # alignment 1
    seq1 = "agcttttcattctgactgcaacgggcaatatgtctctgtgtggattaaaaaaagagtgtctgatagcagcttctgaactg"
    pos1 = 1759
    cig1 = "79S81M80S"
    sap1 = "U00096.2,1760,+,79S81M80S,60,0"
    # alignment 2
    seq2 = "cgttggcggtgcgctgctggagcaactgaagcgtcagcaaagctggctgaagaataaacatatcgacttacgtgtctgcg"
    pos2 = 0
    cig2 = "80M160S"
    sap2 = "U00096.2,1,+,80M160S,60,0"
    # alignment 3
    seq3 = "acttaacgctgctcgtagcgtttaaacaccagttcgccattgctggaggaatcttcatcaaagaagtaaccttcgctatt"
    pos3 = 5680
    cig3 = "160S80M"
    sap3 = "U00096.2,5681,+,160S80M,60,0"

    opt = BwaMemOptions(query_coord_as_primary=query_coord_as_primary, soft_clip_supplementary=True)
    bwa = BwaMem(prefix=e_coli_k12_fasta)
    query = FastxRecord(name="query", sequence=seq1 + seq2 + seq3)

    list_of_recs = bwa.align(opt=opt, queries=[query])
    assert len(list_of_recs) == 1
    assert len(list_of_recs[0]) == 3
    rec1 = list_of_recs[0][0]
    rec2 = list_of_recs[0][1]
    rec3 = list_of_recs[0][2]
    sa1 = f"{sap2};{sap3};"
    sa2 = f"{sap1};{sap3};"
    sa3 = f"{sap1};{sap2};"

    if query_coord_as_primary:  # query coordinate
        # swap rec1 and rec2, rec2 is primary
        tmp = rec1
        rec1 = rec2
        rec2 = tmp
        sa3 = f"{sap2};{sap1};"  # primary order changed

    assert rec1.reference_start == pos1
    assert rec2.reference_start == pos2
    assert rec3.reference_start == pos3

    assert rec1.cigarstring == cig1
    assert rec2.cigarstring == cig2
    assert rec3.cigarstring == cig3

    assert rec1.get_tag("SA") == sa1
    assert rec2.get_tag("SA") == sa2
    assert rec3.get_tag("SA") == sa3


def test_bwamem_align_hard_clip_supplementary(e_coli_k12_fasta: Path) -> None:
    # alignments below are defined based on (1) alignment score, then (2) genomic order
    # alignment 1
    seq1 = "agcttttcattctgactgcaacgggcaatatgtctctgtgtggattaaaaaaagagtgtctgatagcagcttctgaactg"
    pos1 = 1759
    cig1 = "79S81M"
    sap1 = "U00096.2,1760,+,79S81M,60,0"
    # alignment 2
    seq2 = "cgttggcggtgcgctgctggagcaactgaagcgtcagcaaagctggctgaagaataaacatatcgacttacgtgtctgcg"
    pos2 = 0
    cig2 = "80M80H"
    sap2 = "U00096.2,1,+,80M80S,60,0"

    opt = BwaMemOptions(soft_clip_supplementary=False)
    bwa = BwaMem(prefix=e_coli_k12_fasta)
    query = FastxRecord(name="query", sequence=seq1 + seq2)
    list_of_recs = bwa.align(opt=opt, queries=[query])
    assert len(list_of_recs) == 1
    assert len(list_of_recs[0]) == 2
    rec1 = list_of_recs[0][0]
    rec2 = list_of_recs[0][1]

    assert rec1.reference_start == pos1
    assert rec2.reference_start == pos2

    assert rec1.cigarstring == cig1
    assert rec2.cigarstring == cig2

    assert rec1.get_tag("SA") == f"{sap2};"
    assert rec2.get_tag("SA") == f"{sap1};"



"""
Fixtures intended to be shared across multiple files in the tests directory.
"""

from pathlib import Path

import pytest
from pysam import FastxRecord


@pytest.fixture(scope="function")
def temp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp("test_vcf")


@pytest.fixture(scope="session")
def e_coli_k12_fasta() -> Path:
    """The path to the e. Coli K12 reference FASTA"""
    cur_dir = Path(__file__).parent
    fasta: Path = cur_dir / "data" / "e_coli_k12.fasta"
    return fasta


@pytest.fixture(scope="session")
def e_coli_k12_fastx_record() -> FastxRecord:
    """Sequence-only FastxRecord that maps to the forward strand at positino 80 for 80bp"""
    sequence = "gttacctgccgtgagtaaattaaaattttattgacttaggtcactaaatactttaaccaatataggcatagcgcacagac"
    return FastxRecord(name="test", sequence=sequence)



# cython: language_level=3

from pybwa.libbwaindex cimport force_bytes_with
import pytest


def test_force_bytes_with() -> None:
    assert force_bytes_with(None) is None
    assert force_bytes_with(b"foo-bar") == b"foo-bar"
    assert force_bytes_with("foo-bar") == "foo-bar"
    with pytest.raises(TypeError, match="Argument must be a string, bytes or unicode"):
        force_bytes_with(42)



from pathlib import Path
from typing import Optional
from typing import Union

import pysam
import pytest
from fgpyo.sequence import reverse_complement
from pysam import FastxRecord

from pybwa import BwaAln
from pybwa import BwaAlnOptions
from pybwa import BwaIndex


def test_bwaaln_options() -> None:
    # default
    BwaAlnOptions()

    options = BwaAlnOptions(
        max_mismatches=1,
        max_gap_opens=2,
        max_gap_extensions=3,
        min_indel_to_end_distance=4,
        max_occurrences_for_extending_long_deletion=5,
        seed_length=6,
        max_mismatches_in_seed=7,
        mismatch_penalty=8,
        gap_open_penalty=9,
        gap_extension_penalty=10,
        stop_at_max_best_hits=11,
        max_hits=12,
        log_scaled_gap_penalty=False,
        find_all_hits=False,
        with_md=False,
        threads=13,
    )

    assert options.max_mismatches == 1
    assert options.max_gap_opens == 2
    assert options.max_gap_extensions == 3
    assert options.min_indel_to_end_distance == 4
    assert options.max_occurrences_for_extending_long_deletion == 5
    assert options.seed_length == 6
    assert options.max_mismatches_in_seed == 7
    assert options.mismatch_penalty == 8
    assert options.gap_open_penalty == 9
    assert options.gap_extension_penalty == 10
    assert options.stop_at_max_best_hits == 11
    assert options.max_hits == 12
    assert not options.log_scaled_gap_penalty
    assert not options.find_all_hits
    assert not options.with_md
    assert options.threads == 13

    options.log_scaled_gap_penalty = True
    assert options.log_scaled_gap_penalty

    options.find_all_hits = True
    assert options.find_all_hits
    assert options.stop_at_max_best_hits == 0x7FFFFFFF

    options.with_md = True
    assert options.with_md


def test_bwaaln_load_index(e_coli_k12_fasta: Path) -> None:
    BwaAln(prefix=e_coli_k12_fasta)
    BwaAln(index=BwaIndex(e_coli_k12_fasta))
    with pytest.raises(ValueError):
        BwaAln()


def test_bwaaln_no_queries(e_coli_k12_fasta: Path) -> None:
    bwa = BwaAln(prefix=e_coli_k12_fasta)
    assert bwa.align(queries=[]) == []


@pytest.mark.parametrize("query_is_str", [True, False])
@pytest.mark.parametrize("with_quals", [True, False])
def test_bwaaln_basic(
    e_coli_k12_fasta: Path,
    e_coli_k12_fastx_record: FastxRecord,
    query_is_str: bool,
    with_quals: bool,
) -> None:
    """
    Args:
        e_coli_k12_fasta: path to the index FASTA
        e_coli_k12_fastx_record: the FASTX record to align
        query_is_str: true if the query is to be passed a string, false if as a FastxRecord
        with_quals: true to provide base qualities, false otherwise. Only works if query_is_str is
                    false.
    """
    opt = BwaAlnOptions()
    bwa = BwaAln(prefix=e_coli_k12_fasta)

    revcomp_seq = (
        None
        if not e_coli_k12_fastx_record.sequence
        else reverse_complement(e_coli_k12_fastx_record.sequence)
    )
    if with_quals and not query_is_str:
        assert e_coli_k12_fastx_record.sequence is not None
        quality = "J"
        quality += "I" * (len(e_coli_k12_fastx_record.sequence) - 1)
        revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq, quality=quality)
        e_coli_k12_fastx_record = FastxRecord(
            name=e_coli_k12_fastx_record.name,
            sequence=e_coli_k12_fastx_record.sequence,
            quality=quality,
        )
    else:
        revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    queries: Union[list[str], list[FastxRecord]]
    if query_is_str:
        queries = [str(q.sequence) for q in [e_coli_k12_fastx_record, revcomp_record]]
    else:
        queries = [e_coli_k12_fastx_record, revcomp_record]

    recs = bwa.align(opt=opt, queries=queries)
    assert len(recs) == 2
    rec = recs[0]
    assert rec.query_name == "read.0" if query_is_str else "test"
    assert rec.query_sequence == e_coli_k12_fastx_record.sequence
    if e_coli_k12_fastx_record.quality is not None:
        assert rec.query_qualities is not None
        assert (
            pysam.qualities_to_qualitystring(rec.query_qualities) == e_coli_k12_fastx_record.quality
        )
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_forward
    assert rec.cigarstring == "80M"

    rec = recs[1]
    assert rec.query_name == "read.1" if query_is_str else "revcomp"
    assert rec.query_sequence == e_coli_k12_fastx_record.sequence
    if revcomp_record.quality is not None:
        assert rec.query_qualities is not None
        assert pysam.qualities_to_qualitystring(rec.query_qualities) == revcomp_record.quality[::-1]
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_reverse
    assert rec.cigarstring == "80M"

    # NB: XN, XA not generated for these records
    expected_tags = ["NM", "X0", "X1", "XM", "XO", "XG", "MD", "HN"]
    for rec in recs:
        for tag in expected_tags:
            assert rec.has_tag(tag), f"Missing tag {tag} in: {rec}"


def test_bwaaln_multi_hit(e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord) -> None:
    opt = BwaAlnOptions(threads=1, seed_length=10, max_hits=1000)
    bwa = BwaAln(prefix=e_coli_k12_fasta)
    assert e_coli_k12_fastx_record.sequence is not None

    queries = [
        FastxRecord(
            name=e_coli_k12_fastx_record.name, sequence=e_coli_k12_fastx_record.sequence[:10]
        )
    ]

    recs = bwa.align(opt=opt, queries=queries)
    assert len(recs) == 1
    rec = recs[0]
    assert rec.reference_start == 4074272
    assert rec.query_name == "test"
    assert queries[0].sequence is not None
    assert rec.query_sequence == reverse_complement(queries[0].sequence)
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.is_reverse
    assert rec.cigarstring == "10M"
    assert rec.has_tag("XA")
    assert rec.has_tag("HN")
    assert rec.get_tag("HN") == 417
    xa_values = str(rec.get_tag("XA")).split(";")[:-1]
    assert len(xa_values) == 416


def test_bwaaln_multi_hit_split_cigar(
    e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord
) -> None:
    opt = BwaAlnOptions(
        threads=1,
        seed_length=5,
        max_hits=10000,
        max_gap_opens=1,
        max_gap_extensions=3,
        gap_open_penalty=1,
    )
    bwa = BwaAln(prefix=e_coli_k12_fasta)
    assert e_coli_k12_fastx_record.sequence is not None

    queries = [
        FastxRecord(
            name=e_coli_k12_fastx_record.name,
            sequence=e_coli_k12_fastx_record.sequence[:5]
            + "AAA"
            + e_coli_k12_fastx_record.sequence[5:10],
        )
    ]

    recs = bwa.align(opt=opt, queries=queries)
    assert len(recs) == 1
    rec = recs[0]
    assert rec.reference_start == 465895
    assert rec.query_name == "test"
    assert queries[0].sequence is not None
    assert rec.query_sequence == reverse_complement(queries[0].sequence)
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.is_reverse
    assert rec.cigarstring == "5M1I7M"
    assert rec.has_tag("XA")
    assert rec.has_tag("HN")
    assert rec.get_tag("HN") == 8
    xa_values = str(rec.get_tag("XA")).split(";")[:-1]
    assert len(xa_values) == 7, xa_values
    # check that we have one XA value with an indel
    assert any(v.split(",")[2] == "4M1I8M" for v in xa_values)


def test_bwa_unmapped(
    e_coli_k12_fasta: Path,
) -> None:
    opt = BwaAlnOptions()
    bwa = BwaAln(prefix=e_coli_k12_fasta)
    recs = bwa.align(opt=opt, queries=["GATTACAGATTACAGATTACAGATTACA"])
    assert len(recs) == 1
    rec = recs[0]
    assert rec.is_unmapped


def test_bwaaln_threading(e_coli_k12_fasta: Path, e_coli_k12_fastx_record: FastxRecord) -> None:
    opt = BwaAlnOptions(threads=2)
    bwa = BwaAln(prefix=e_coli_k12_fasta)
    revcomp_seq = (
        None
        if not e_coli_k12_fastx_record.sequence
        else reverse_complement(e_coli_k12_fastx_record.sequence)
    )
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    queries = [e_coli_k12_fastx_record if i % 2 == 0 else revcomp_record for i in range(100)]
    recs = bwa.align(opt=opt, queries=queries)
    assert len(recs) == len(queries)
    for i, rec in enumerate(recs):
        if i % 2 == 0:
            assert rec.query_name == "test"
            assert rec.is_forward
        else:
            assert rec.query_name == "revcomp"
            assert rec.is_reverse
        assert not rec.is_paired
        assert not rec.is_read1
        assert not rec.is_read2
        assert rec.reference_start == 80
        assert rec.cigarstring == "80M"


@pytest.mark.parametrize(
    "max_gap_extensions,is_mapped,deletion_length",
    [
        (None, True, 4),  # 4bp deletions allowed (max diff is 4)
        (None, False, 5),  # 5bp deletions disallowed (max diff is 4)
        (-1, True, 1),  # 1bp deletions allowed
        (-1, False, 2),  # 2bp deletions disallowed
        (0, True, 1),  # 1bp deletions allowed
        (0, False, 2),  # 2bp deletions disallowed
        (4, True, 5),  # 5bp deletion allowed (1 open + 4 extensions)
        (4, False, 6),  # 6bp deletion disallowed (1 open + 4 extensions)
    ],
)
def test_bwaaln_with_deletion(
    e_coli_k12_fasta: Path,
    e_coli_k12_fastx_record: FastxRecord,
    deletion_length: int,
    max_gap_extensions: Optional[int],
    is_mapped: bool,
) -> None:
    bwa = BwaAln(prefix=e_coli_k12_fasta)

    # Create a deletion
    assert e_coli_k12_fastx_record.sequence is not None
    sequence = (
        e_coli_k12_fastx_record.sequence[:40]
        + e_coli_k12_fastx_record.sequence[40 + deletion_length :]
    )
    e_coli_k12_fastx_record = FastxRecord(name=e_coli_k12_fastx_record.name, sequence=sequence)

    opt = BwaAlnOptions(max_gap_extensions=max_gap_extensions)
    recs = bwa.align(opt=opt, queries=[e_coli_k12_fastx_record])
    assert len(recs) == 1
    rec = recs[0]
    assert rec.query_name == "test"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2

    if is_mapped:
        assert rec.reference_start == 80
        assert rec.is_forward
        assert rec.cigarstring == f"40M{deletion_length}D{40-deletion_length}M"
    else:
        assert rec.reference_start == -1, e_coli_k12_fastx_record.sequence
        assert rec.is_unmapped


def test_bwa_aln_map_one_multi_mapped_max_hits_one(e_coli_k12_fasta: Path) -> None:
    """Tests that a query that returns too many hits (>max_hits) returns the number of hits but
    not the list of hits themselves."""
    opt = BwaAlnOptions(
        threads=2,
        max_hits=1,
        max_mismatches=3,
        max_mismatches_in_seed=3,
        max_gap_opens=0,
        max_gap_extensions=-1,
        min_indel_to_end_distance=3,
        seed_length=20,
        find_all_hits=True,
        with_md=True,
    )
    bwa = BwaAln(prefix=e_coli_k12_fasta)
    queries = [FastxRecord(name="NA", sequence="TTTTT")]
    recs = bwa.align(opt=opt, queries=queries)
    assert len(recs) == 1
    rec = recs[0]
    assert rec.has_tag("HN"), str(rec)
    assert rec.get_tag("HN") == 3269888
